/* Sidebar header background color */
.wy-side-nav-search,
.wy-side-nav-title {
    background-color: #004238;
    /* Green background color */
    color: #ffffff;
    /* White text color */
}

/* Change the background color of the documentation container */
.wy-nav-content-wrap {
    background-color: #F7F8D8;
    /* Light green background color */
}

/* Change the background color of the documentation content */
.wy-nav-content {
    background-color: #F7F8D8;
    /* Light green background color */
}